import sys
import socket
import os
import time
import neotermcolor
from neotermcolor import colored, cprint
from datetime import datetime
from rapidtables import (format_table, FORMAT_GENERATOR, FORMAT_GENERATOR_COLS,
                         MULTILINE_ALLOW)
from collections import OrderedDict

from .sharedobj import current_command, common


def ok():
    print(colored('OK', color='green'))


def debug(s):
    print(colored(str(s), color='grey'))


def warn(s, delay=False):
    print(colored(str(s), color='yellow', attrs='bold'))


def err(e, delay=False):
    print(colored(str(e), color='red'))
    if delay:
        import getch
        if getch.getch() == '\x03':
            raise KeyboardInterrupt


def prepare_time(t):
    if t is None:
        return None
    import dateutil.parser
    try:
        return float(t)
    except:
        return dateutil.parser.parse(t).timestamp()


def print_tb(force=False, delay=False):
    if force:
        import traceback
        err(traceback.format_exc())
    else:
        err('FAILED')
    if delay:
        import getch
        if getch.getch() == '\x03':
            raise KeyboardInterrupt


def can_colorize():
    return os.getenv('ANSI_COLORS_DISABLED') is None and (
        not neotermcolor.tty_aware or neotermcolor._isatty)


def set_file_lock(name):
    lock_file = f'{common.dir_eva}/var/{name}.lock'
    if os.path.exists(lock_file):
        raise RuntimeError
    else:
        with open(lock_file, 'w'):
            pass


def remove_file_lock(name):
    lock_file = f'{common.dir_eva}/var/{name}.lock'
    try:
        os.unlink(lock_file)
    except FileNotFoundError:
        pass


def get_my_ip():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        s.connect(('255.255.255.255', 0))
        ip = s.getsockname()[0]
        try:
            s.close()
        except:
            pass
    except:
        ip = None
    return ip


TIME_ORD = {
    'created': 0,
    'accepted': 1,
    'pending': 0b10,
    'running': 0b1000,
    'completed': 0b1111,
    'failed': 0b10000000,
    'canceled': 0b10000001,
    'terminated': 0b10000010,
}

ACTION_STATUS_COLOR = {
    'completed': 'green',
    'canceled': 'grey',
    'terminated': 'yellow',
    'failed': 'red'
}


def xc(cmd, ps='working', verbose=False):
    import subprocess

    is_tty = can_colorize()

    CAROUSEL = ['-', '\\', '|', '/']

    def bs():
        print('\033[D', end='')

    def print_st(st, color=None):
        data = st.read()
        if data:
            try:
                cprint(data.decode(), color=color)
            except:
                cprint(data, color=color)

    if verbose:
        print(ps)
        p = subprocess.run(cmd, shell=True)
    else:
        print(ps, end='  ' if is_tty else '... ', flush=True)
        p = subprocess.Popen(cmd,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        c_idx = 0
        while p.poll() is None:
            if is_tty:
                bs()
                print(CAROUSEL[c_idx], end='', flush=True)
                c_idx += 1
                if c_idx == len(CAROUSEL):
                    c_idx = 0
            time.sleep(0.1)
        if is_tty:
            bs()
        if p.returncode == 0:
            cprint('OK', color='green', flush=True)
        else:
            cprint('FAILED!', color='red', attrs=['bold'], flush=True)
            print_st(p.stdout)
            print_st(p.stderr, color='red')
    if p.returncode != 0:
        raise RuntimeError(f'process exited with the code {p.returncode}')


def exec_cmd(cmd, args, search_in='bin', search_system=True):
    c = f'{common.dir_eva}/{search_in}/{cmd}'
    if not os.path.isfile(c) and search_system:
        import shutil
        c = shutil.which(cmd)
    if not c:
        raise RuntimeError(f'{cmd} not found')
    code = os.system(f'{c} {args}')
    if code:
        raise RuntimeError(f'{cmd} failed with code {code}')


def print_action_result(result):
    import uuid
    result['uuid'] = str(uuid.UUID(bytes=result['uuid']))
    status = result['status']
    time = result['time']
    times = [v for _, v in time.items()]
    if status in ['completed', 'failed', 'terminated']:
        result['elapsed'] = '{:.6f}'.format(max(times) - min(times))
    if time:
        time_data = sorted([{
            'n': k,
            'v': datetime.fromtimestamp(v, common.TZ).isoformat()
        } for k, v in time.items()],
                           key=lambda k: TIME_ORD.get(k['n'], 10))
        result['time'] = format_table(time_data, generate_header=False)
    params = result['params']
    if params:
        params_data = [{'n': k, 'v': v} for k, v in params.items()]
        result['params'] = format_table(params_data, generate_header=False)
    out = result.pop('out')
    err = result.pop('err')
    data = sorted([{
        'n': k,
        'v': v
    } for k, v in result.items()],
                  key=lambda k: k['n'])
    rows = format_table(
        data,
        fmt=FORMAT_GENERATOR_COLS,
        generate_header=False,
        multiline=MULTILINE_ALLOW,
    )
    spacer = '  '
    for r in rows:
        print(colored(r[0], color='blue') + spacer, end='')
        if r[0].startswith('status '):
            print(colored(r[1], color=ACTION_STATUS_COLOR.get(status)))
        else:
            print(r[1])
    if out is not None:
        print('--- OUT ---')
        print(out)
    if err is not None:
        print('--- ERR ---')
        print(colored(err, color='red'))
    print()


def print_result(data, need_header=True, name_value=False, cols=None):
    if current_command.json:
        from pygments import highlight, lexers, formatters
        import json
        j = json.dumps(data, indent=4, sort_keys=True)
        if can_colorize():
            j = highlight(j, lexers.JsonLexer(), formatters.TerminalFormatter())
        print(j)
        return
    elif data:
        if name_value:
            if name_value is True:
                nn = 'field'
                vn = 'value'
            else:
                nn = name_value[0]
                vn = name_value[1]
            data = sorted([{
                nn: k,
                vn: v
            } for k, v in data.items()],
                          key=lambda k: k[nn])
            if need_header:
                header, rows = format_table(data, fmt=FORMAT_GENERATOR)
                print(colored(header, color='blue'))
                print(colored('-' * len(header), color='grey'))
                for r in rows:
                    print(r)
            else:
                rows = format_table(data,
                                    fmt=FORMAT_GENERATOR_COLS,
                                    generate_header=False)
                spacer = '  '
                for r in rows:
                    print(colored(r[0], color='blue') + spacer, end='')
                    print(r[1])
        else:
            if cols:
                col_rules = {}
                for i, c in enumerate(cols):
                    if '|' in c:
                        r = c.split('|')
                        rules = {'_': r[0]}
                        col_rules[i] = rules
                        for rule in r[1:]:
                            k, v = rule.split('=', maxsplit=1)
                            rules[k] = v
                    else:
                        col_rules[i] = {}
                formatted_data = []
                for d in data:
                    od = OrderedDict()
                    for i, c in enumerate(cols):
                        rules = col_rules[i]
                        src = rules.get('_', c)
                        val = d.get(src, '')
                        fmt = rules.get('f')
                        if val != '' and val is not None:
                            try:
                                if fmt is not None:
                                    if fmt == 'time':
                                        val = datetime.fromtimestamp(
                                            val, common.TZ).isoformat()
                                    elif fmt.startswith('time:'):
                                        zone = fmt.split(':', 1)[-1]
                                        import pytz
                                        val = datetime.fromtimestamp(
                                            val,
                                            pytz.timezone(zone)).isoformat()
                                    elif fmt.startswith('round:'):
                                        digits = int(fmt.split(':', 1)[-1])
                                        if digits > 0:
                                            val = round(val, digits)
                                        else:
                                            val = int(val)
                            except:
                                pass
                        od[rules.get('n', c)] = val
                    formatted_data.append(od)
                header, rows = format_table(formatted_data,
                                            fmt=FORMAT_GENERATOR)
            else:
                if data and not isinstance(data[0], dict):
                    data = [{'value': d} for d in data]
                header, rows = format_table(data, fmt=FORMAT_GENERATOR)
            if need_header:
                print(colored(header, color='blue'))
                print(colored('-' * len(header), color='grey'))
            for r in rows:
                print(r)
    print()


def edit_file(fname):
    from pathlib import Path
    suffix = Path(fname).suffix
    fname = f'{common.dir_eva}/{fname}'
    editor = os.getenv('EDITOR', 'vi')
    while True:
        code = os.system(f'{editor} "{fname}"')
        if code:
            err(f'editor exited with code {code}')
            break
        try:
            if suffix in ['.yaml', '.yml']:
                import yaml
                with open(fname) as fh:
                    yaml.safe_load(fh.read())
            elif suffix == '.json':
                import json
                with open(fname) as fh:
                    json.loads(fh.read())
            elif suffix == '.py':
                with open(fname) as fh:
                    compile(fh.read(), fname, 'exec')
            break
        except:
            print_tb(force=True, delay=True)
            continue


def edit_config(value, ss, deploy_fn, initial=False):
    import yaml
    import tempfile
    from hashlib import sha256
    from pathlib import Path
    editor = os.getenv('EDITOR', 'vi')
    fname = sha256(ss.encode()).hexdigest()
    tmpfile = Path(f'{tempfile.gettempdir()}/{fname}.tmp.yml')
    if isinstance(value, str):
        tmpfile.write_text(value)
    else:
        tmpfile.write_text(yaml.dump(value, default_flow_style=False))
    try:
        while True:
            code = os.system(f'{editor} {tmpfile}')
            if code:
                err(f'editor exited with code {code}')
                break
            try:
                data = yaml.safe_load(tmpfile.read_text())
            except:
                print_tb(force=True, delay=True)
                continue
            if data == value and not initial:
                break
            else:
                try:
                    deploy_fn(data)
                    break
                except Exception as e:
                    err(e, delay=True)
                    continue
    finally:
        try:
            tmpfile.unlink()
        except FileNotFoundError:
            pass


def read_file(fname=None):
    if fname is None or fname == '-':
        if sys.stdin.isatty():
            print(
                'Copy/paste or type data, press Ctrl-D to end, Ctrl-C to abort')
        return sys.stdin.buffer.read()
    else:
        with open(os.path.expanduser(fname), 'rb') as fh:
            return fh.read()


def write_file(fname, content, mode='w'):
    with open(os.path.expanduser(fname), mode) as fh:
        fh.write(content)


def format_value(value, advanced=False):
    if value.startswith('!'):
        return value[1:]
    elif advanced and ',' in value:
        return [format_value(v) for v in value.split(',')]
    else:
        if value == 'null':
            return None
        elif value == 'true':
            return True
        elif value == 'false':
            return False
        try:
            return int(value)
        except:
            try:
                return float(value)
            except:
                return value


def safe_print(val, extra=0):
    width, height = get_term_size()
    print(val[:width + extra])


def get_term_size():
    width, height = os.get_terminal_size(0)
    if width == 0 or height == 0:
        return (80, 25)
    else:
        return (width, height)


def get_node_svc_info():
    import json
    return json.loads(
        os.popen(f'{common.dir_eva}/svc/eva-node --mode info').read())


def get_arch():
    import platform
    return platform.machine()



import os
import sys
import time
from functools import partial

import elbus
import msgpack
from neotermcolor import colored
from rapidtables import format_table, FORMAT_GENERATOR, FORMAT_GENERATOR_COLS

from .tools import print_result, ok, edit_config, read_file, write_file
from .tools import print_action_result, format_value, prepare_time, err, warn
from .tools import get_node_svc_info, edit_file, get_term_size, exec_cmd, xc
from .tools import get_my_ip
from .client import call_rpc, DEFAULT_REPL_SERVICE
from .sharedobj import common, current_command

from . import DEFAULT_REPOSITORY_URL


def eva_control_c(c, pfx=''):
    cmd = f'{pfx}{common.dir_eva}/sbin/eva-control {c}'
    os.system(cmd)


class CLI:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for vc in [
                'build', 'list', 'edit', 'config', 'add', 'remove', 'update',
                'mirror-update', 'mirror-set'
        ]:
            setattr(self, f'venv_{vc.replace("-", "_")}',
                    partial(self.venv_cmd, vc))

    def venv_cmd(self, vc, **kwargs):
        args = '--verbose ' if current_command.debug else ''
        args += vc
        args += ' ' + ' '.join(kwargs.get('modules', []))
        if 'mirror_url' in kwargs:
            args += f' "{kwargs["mirror_url"]}"'
        if kwargs.get('rebuild'):
            args += ' --rebuild'
        if kwargs.get('from_scratch'):
            args += ' --from-scratch'
        if 'dest' in kwargs:
            args += f' --dest "{kwargs["dest"]}"'
        exec_cmd('venvmgr', args, search_in='sbin', search_system=False)

    def test(self):
        data = call_rpc('test')
        print_result(data, need_header=False, name_value=True)

    def save(self):
        call_rpc('save')
        ok()

    def server_start(self):
        eva_control_c('start')
        time.sleep(1)

    def server_stop(self):
        eva_control_c('stop')
        time.sleep(1)

    def server_restart(self):
        eva_control_c('restart')
        time.sleep(1)

    def server_reload(self):
        until = time.time() + current_command.timeout
        call_rpc('core.shutdown')
        while True:
            time.sleep(1)
            try:
                call_rpc('test')
                break
            except:
                pass
            if time.time() > until:
                raise TimeoutError
        ok()

    def server_launch(self):
        eva_control_c('launch', 'VERBOSE=1 ')

    def server_status(self):
        eva_control_c('status')

    def cloud_deploy(self, file, config_var, config, test):
        args = (f'cloud deploy --timeout {current_command.timeout}'
                f' --connection-path "{common.elbus_path}" "{file}"')
        if config_var:
            for c in config_var:
                args += f' --config-var "{c}"'
        if config:
            args += f' --config "{config}"'
        if test:
            args += ' --test'
        exec_cmd('eva-cloud-manager', args)

    def cloud_undeploy(self, file, config_var, config, test):
        args = (f'cloud undeploy --timeout {current_command.timeout}'
                f' --connection-path "{common.elbus_path}" "{file}"')
        if config_var:
            for c in config_var:
                args += f' --config-var "{c}"'
        if config:
            args += f' --config "{config}"'
        if test:
            args += ' --test'
        exec_cmd('eva-cloud-manager', args)

    def cloud_update(self, nodes, check_timeout, all, info_only, yes):
        if not nodes and not all:
            raise RuntimeError('specify either nodes to update or --all')
        args = (f'cloud update --timeout {current_command.timeout}'
                f' --connection-path "{common.elbus_path}"'
                f' --check-timeout "{check_timeout}"')
        args += ' ' + ' '.join([f'"{n}"' for n in nodes])
        if all:
            args += ' --all'
        if info_only:
            args += ' --info-only'
        if yes:
            args += ' --YES'
        exec_cmd('eva-cloud-manager', args)

    def update(self, download_timeout, repository_url, yes, info_only, test):
        args = f'node update --timeout {current_command.timeout}'
        if repository_url:
            args = + f' --repository-url "{repository_url}"'
        if download_timeout:
            args += f' --download-timeout {download_timeout}'
        if yes:
            args += ' --YES'
        if info_only:
            args += ' --info-only'
        if test:
            args += ' --test'
        old_ver = get_node_svc_info()
        exec_cmd('eva-cloud-manager', args)
        if not info_only:
            if old_ver != get_node_svc_info():
                print('Update completed', end='')
                if common.interactive:
                    print('. Now exit EVA shell and log in back')
                else:
                    print()

    def registry_manage(self, yedb_args=None):
        cmd = f'{common.dir_eva}/venv/bin/yedb'
        args = '' if yedb_args is None else ' ' + yedb_args
        if not os.path.exists(cmd):
            import shutil
            cmd = shutil.which('yedb')
            if not cmd:
                raise RuntimeError('YEDB CLI not found')
        try:
            data = call_rpc('test')
            system_name = data.get('system_name', '')
            yedb_socket = f'elbus://{common.elbus_path}:eva.registry'
            os.system(f'env "YEDB_PS=registry:{system_name}" "{cmd}" '
                      f'"{yedb_socket}"{args}')
        except:
            import getch
            warn('The node is offline, starting '
                 'the registry manager in offline mode')
            warn('The registry database will be '
                 'LOCKED until the registry shell exit')
            warn('Press any key to continue, Ctrl-C to abort...')
            if getch.getch() != '\x03':
                os.system(f'env YEDB_PS=registry "{cmd}"'
                          f' "{common.dir_eva}/runtime/registry"{args}')

    def edit(self, fname, offline=False):
        if fname.startswith('config'):
            self.registry_manage(yedb_args=f'edit eva/"{fname}"')
            if fname != 'config/python-venv':
                print('restart or reload the node server to apply changes')
        else:
            edit_file(fname)

    def log_purge(self):
        call_rpc('log.purge')
        ok()

    def log_get(self, level, time, limit, module, regex, full):

        def log_record_color(level):
            if level == 'TRACE':
                return dict(color='grey', attrs='dark')
            elif level == 'DEBUG':
                return dict(color='grey')
            elif level == 'WARN':
                return dict(color='yellow', attrs='bold')
            elif level == 'ERROR':
                return dict(color='red')
            return {}

        width, height = get_term_size()

        params = {
            'level': level,
            'time': time,
            'limit': limit if limit else height - 3,
            'module': module,
            'rx': regex
        }
        records = call_rpc('log.get', params)
        if current_command.json:
            print_result(records)
        elif records:
            data = []
            for r in records:
                data.append(
                    dict(time=r['dt'],
                         host=r['h'],
                         module=r['mod'],
                         level=r['lvl'].upper(),
                         message=r['msg']))
            header, rows = format_table(data, fmt=FORMAT_GENERATOR)
            if not full:
                header = header[:width]
            print(colored(header, color='blue'))
            print(colored('-' * len(header), color='grey'))
            for row, record in zip(rows, data):
                print(
                    colored(row if full else row[:width],
                            **log_record_color(record['level'])))

    def acl_list(self, acl_svc):
        data = call_rpc('acl.list', target=acl_svc)
        print_result(data, cols=['id', 'admin'])

    def acl_edit(self, i, acl_svc):

        def deploy(cfg, i):
            call_rpc('acl.deploy', dict(acls=[cfg]), target=acl_svc)
            print(f'ACL re-deployed: {i}')
            print()

        config = call_rpc('acl.get_config', dict(i=i), target=acl_svc)
        edit_config(config, f'acl|{i}|config', deploy_fn=partial(deploy, i=i))

    def acl_export(self, i, acl_svc, output=None):
        c = 0
        configs = []
        for acl in call_rpc('acl.list', target=acl_svc):
            name = acl['id']
            if (i.startswith('*') and name.endswith(i[1:])) or \
                (i.endswith('*') and name.startswith(i[:-1])) or \
                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
                i == name:
                c += 1
                configs.append(acl)
        result = dict(acls=configs)
        if current_command.json:
            print_result(result)
        else:
            import yaml
            dump = yaml.dump(result, default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} ACL(s) exported')
                print()

    def acl_deploy(self, acl_svc, file=None):
        import yaml
        acls = yaml.safe_load(read_file(file).decode()).pop('acls')
        call_rpc('acl.deploy', dict(acls=acls), target=acl_svc)
        print(f'{len(acls)} ACL(s) deployed')
        print()

    def acl_undeploy(self, acl_svc, file=None):
        import yaml
        acls = yaml.safe_load(read_file(file).decode()).pop('acls')
        call_rpc('acl.undeploy', dict(acls=acls), target=acl_svc)
        print(f'{len(acls)} ACL(s) undeployed')
        print()

    def acl_create(self, i, acl_svc):
        call_rpc('acl.deploy', dict(acls=[dict(id=i)]), target=acl_svc)
        ok()

    def acl_destroy(self, i, acl_svc):
        call_rpc('acl.destroy', dict(i=i), target=acl_svc)
        ok()

    def key_list(self, auth_svc):
        data = call_rpc('key.list', target=auth_svc)
        if not current_command.json:
            for d in data:
                d['acls'] = ', '.join(d['acls'])
        print_result(data, cols=['id', 'acls'])

    def key_edit(self, i, auth_svc):

        def deploy(cfg, i):
            call_rpc('key.deploy', dict(keys=[cfg]), target=auth_svc)
            print(f'API key re-deployed: {i}')
            print()

        config = call_rpc('key.get_config', dict(i=i), target=auth_svc)
        edit_config(config, f'key|{i}|config', deploy_fn=partial(deploy, i=i))

    def key_export(self, i, auth_svc, output=None):
        c = 0
        configs = []
        for key in call_rpc('key.list', target=auth_svc):
            name = key['id']
            if (i.startswith('*') and name.endswith(i[1:])) or \
                (i.endswith('*') and name.startswith(i[:-1])) or \
                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
                i == name:
                c += 1
                configs.append(key)
        result = dict(keys=configs)
        if current_command.json:
            print_result(result)
        else:
            import yaml
            dump = yaml.dump(result, default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} API key(s) exported')
                print()

    def key_deploy(self, auth_svc, file=None):
        import yaml
        keys = yaml.safe_load(read_file(file).decode()).pop('keys')
        call_rpc('key.deploy', dict(keys=keys), target=auth_svc)
        print(f'{len(keys)} API key(s) deployed')
        print()

    def key_undeploy(self, auth_svc, file=None):
        import yaml
        keys = yaml.safe_load(read_file(file).decode()).pop('keys')
        call_rpc('key.undeploy', dict(keys=keys), target=auth_svc)
        print(f'{len(keys)} API key(s) undeployed')
        print()

    def key_create(self, i, auth_svc):
        import random
        import string
        symbols = string.ascii_letters + '0123456789'
        k = ''.join(random.choice(symbols) for i in range(16))
        call_rpc('key.deploy', dict(keys=[dict(id=i, key=k)]), target=auth_svc)
        data = call_rpc('key.get', dict(i=i), target=auth_svc)
        print_result([data], cols=['id', 'key'])

    def key_destroy(self, i, auth_svc):
        call_rpc('key.destroy', dict(i=i), target=auth_svc)
        ok()

    def key_regenerate(self, i, auth_svc):
        data = call_rpc('key.regenerate', dict(i=i), target=auth_svc)
        print_result([data], cols=['id', 'key'])

    def key_get(self, i, auth_svc):
        data = call_rpc('key.get_config', dict(i=i), target=auth_svc)
        if not current_command.json:
            data['acls'] = ', '.join(data['acls'])
        print_result([data], cols=['id', 'key', 'acls'])

    def user_list(self, auth_svc):
        data = call_rpc('user.list', target=auth_svc)
        if not current_command.json:
            for d in data:
                d['acls'] = ', '.join(d['acls'])
        print_result(data, cols=['login', 'acls'])

    def user_edit(self, i, auth_svc):

        def deploy(cfg, i):
            call_rpc('user.deploy', dict(users=[cfg]), target=auth_svc)
            print(f'user re-deployed: {i}')
            print()

        config = call_rpc('user.get_config', dict(i=i), target=auth_svc)
        edit_config(config, f'user|{i}|config', deploy_fn=partial(deploy, i=i))

    def user_export(self, i, auth_svc, output=None):
        c = 0
        configs = []
        for user in call_rpc('user.list', target=auth_svc):
            name = user['login']
            if (i.startswith('*') and name.endswith(i[1:])) or \
                (i.endswith('*') and name.startswith(i[:-1])) or \
                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
                i == name:
                c += 1
                configs.append(user)
        result = dict(users=configs)
        if current_command.json:
            print_result(result)
        else:
            import yaml
            dump = yaml.dump(result, default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} user(s) exported')
                print()

    def user_deploy(self, auth_svc, file=None):
        import yaml
        users = yaml.safe_load(read_file(file).decode()).pop('users')
        call_rpc('user.deploy', dict(users=users), target=auth_svc)
        print(f'{len(users)} user(s) deployed')
        print()

    def user_undeploy(self, auth_svc, file=None):
        import yaml
        users = yaml.safe_load(read_file(file).decode()).pop('users')
        call_rpc('user.undeploy', dict(users=users), target=auth_svc)
        print(f'{len(users)} user(s) undeployed')
        print()

    def user_create(self, i, auth_svc):
        import pwinput
        from hashlib import sha256
        password = pwinput.pwinput()
        call_rpc(
            'user.deploy',
            dict(users=[
                dict(login=i, password=sha256(password.encode()).hexdigest())
            ]),
            target=auth_svc)
        ok()

    def user_destroy(self, i, auth_svc):
        call_rpc('user.destroy', dict(i=i), target=auth_svc)
        ok()

    def user_password(self, i, auth_svc):
        import pwinput
        password = pwinput.pwinput()
        data = call_rpc('user.set_password',
                        dict(i=i, password=password),
                        target=auth_svc)
        ok()

    def user_get(self, i, auth_svc):
        data = call_rpc('user.get_config', dict(i=i), target=auth_svc)
        if not current_command.json:
            data['acls'] = ', '.join(data['acls'])
        print_result([data], cols=['login', 'acls'])

    def svc_list(self):
        data = call_rpc('svc.list')
        print_result(data, cols=['id', 'status', 'pid', 'launcher'])

    def svc_export(self, i, output=None):
        c = 0
        configs = []
        for svc in call_rpc('svc.list'):
            name = svc['id']
            if (i.startswith('*') and name.endswith(i[1:])) or \
                (i.endswith('*') and name.startswith(i[:-1])) or \
                (i.startswith('*') and i.endswith('*') and i[1:-1] in name) or \
                i == name:
                c += 1
                configs.append(
                    dict(id=name,
                         params=call_rpc('svc.get_params', dict(i=name))))
        result = dict(svcs=configs)
        if current_command.json:
            print_result(result)
        else:
            import yaml
            dump = yaml.dump(dict(svcs=configs), default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} service(s) exported')
                print()

    def svc_deploy(self, file=None):
        import yaml
        svcs = yaml.safe_load(read_file(file).decode()).pop('svcs')
        call_rpc('svc.deploy', dict(svcs=svcs))
        print(f'{len(svcs)} service(s) deployed')
        print()

    def svc_undeploy(self, file=None):
        import yaml
        svcs = yaml.safe_load(read_file(file).decode()).pop('svcs')
        call_rpc('svc.undeploy', dict(svcs=svcs))
        print(f'{len(svcs)} service(s) undeployed')
        print()

    def svc_restart(self, i):
        call_rpc('svc.restart', dict(i=i))
        ok()

    def svc_destroy(self, i):
        call_rpc('svc.undeploy', dict(svcs=[i]))
        ok()

    def svc_purge(self, i):
        call_rpc('svc.purge', dict(svcs=[i]))
        ok()

    def svc_call(self, i, file, method, params):
        if file:
            import yaml
            payload = yaml.safe_load(read_file(file))
        else:
            payload = {}
            for p in params:
                n, v = p.split('=', 1)
                payload[n] = format_value(v, advanced=True)
        result = call_rpc(method, payload if payload else None, target=i)
        if current_command.json or isinstance(result, list):
            print_result(result)
        elif isinstance(result, dict):
            print_result(result, name_value=True)
        elif result is None:
            ok()
        else:
            print(result)
            print()

    def svc_edit(self, i):

        def deploy(cfg, i):
            call_rpc('svc.deploy', dict(svcs=[dict(id=i, params=cfg)]))
            print(f'service re-deployed: {i}')
            print()

        config = call_rpc('svc.get_params', dict(i=i))
        edit_config(config, f'svc|{i}|config', deploy_fn=partial(deploy, i=i))

    def svc_create(self, i, f):

        def deploy(cfg, i):
            call_rpc('svc.deploy', dict(svcs=cfg))
            print(f'service deployed: {i}')
            print()

        lines = read_file(f).decode().split('\n')
        tpl = (' ' * 4) + ('\n' + ' ' * 4).join(lines).rstrip()
        config = f'- id: {i}\n  params:\n{tpl}'
        if sys.stdin.isatty():
            edit_config(config,
                        f'svc|{i}|config',
                        deploy_fn=partial(deploy, i=i))
        else:
            import yaml
            deploy(yaml.safe_load(config), i)

    def svc_test(self, i):
        call_rpc('test', target=i)
        ok()

    def svc_info(self, i):
        data = call_rpc('info', target=i)
        if not current_command.json:
            try:
                del data['methods']
            except KeyError:
                pass
        print_result(data, need_header=False, name_value=True)

    def broker_test(self):
        call_rpc('test', target='.broker')
        ok()

    def broker_info(self):
        data = call_rpc('info', target='.broker')
        print_result(data, need_header=False, name_value=True)

    def broker_stats(self):
        data = call_rpc('stats', target='.broker')
        print_result(data, need_header=False, name_value=True)

    def broker_client_list(self):
        data = call_rpc('client.list', target='.broker')
        if current_command.json:
            print_result(data)
        else:
            print_result(
                [d for d in data['clients'] if d['name'] != common.elbus_name])

    def action_exec(self, i, status, value, priority, wait):
        params = dict(i=i, wait=wait)
        if priority is not None:
            params['priority'] = int(priority)
        action_params = {'status': int(status)}
        if value is not None:
            action_params['value'] = format_value(value)
        params['params'] = action_params
        result = call_rpc('action', params)
        if current_command.json:
            import uuid
            result['uuid'] = str(uuid.UUID(bytes=result['uuid']))
            print_result(result)
        else:
            print_action_result(result)

    def action_run(self, i, arg, kwarg, priority, wait):
        params = dict(i=i, wait=wait)
        if priority is not None:
            params['priority'] = int(priority)
        action_params = {}
        if arg:
            action_params['args'] = [format_value(v) for v in arg]
        if kwarg:
            kw = {}
            for k in kwarg:
                n, v = k.split('=', 1)
                kw[n] = format_value(v)
            action_params['kwargs'] = kw
        params['params'] = action_params
        result = call_rpc('run', params)
        if current_command.json:
            import uuid
            result['uuid'] = str(uuid.UUID(bytes=result['uuid']))
            print_result(result)
        else:
            print_action_result(result)

    def action_toggle(self, i, priority, wait):
        params = dict(i=i, wait=wait)
        if priority is not None:
            params['priority'] = int(priority)
        result = call_rpc('action.toggle', params)
        if current_command.json:
            import uuid
            result['uuid'] = str(uuid.UUID(bytes=result['uuid']))
            print_result(result)
        else:
            print_action_result(result)

    def action_terminate(self, u):
        import uuid
        call_rpc('action.terminate', dict(u=uuid.UUID(u).bytes))
        ok()

    def action_kill(self, i):
        call_rpc('action.kill', dict(i=i))
        ok()

    def action_result(self, u):
        import uuid
        result = call_rpc('action.result', dict(u=uuid.UUID(u).bytes))
        if current_command.json:
            import uuid
            result['uuid'] = str(uuid.UUID(bytes=result['uuid']))
            print_result(result)
        else:
            print_action_result(result)

    def action_list(self, oid, status_query, svc, time, limit):
        import uuid
        from .tools import ACTION_STATUS_COLOR
        params = dict(i=oid, sq=status_query, svc=svc, time=time)
        if limit is not None:
            params['limit'] = limit
        result = call_rpc('action.list', params)
        if current_command.json:
            for r in result:
                r['uuid'] = str(uuid.UUID(bytes=r['uuid']))
            print_result(result)
        elif result:
            data = []
            for r in result:
                if r['status'] in ['completed', 'failed', 'terminated']:
                    times = [v for _, v in r['time'].items()]
                    elapsed = '{:.6f}'.format(max(times) - min(times))
                else:
                    elapsed = ''
                data.append({
                    'uuid': str(uuid.UUID(bytes=r['uuid'])),
                    'oid': r['oid'],
                    'status': r['status'],
                    'elapsed': elapsed,
                    'node': r['node'],
                    'svc': r['svc']
                })
            spacer = '  '
            header, rows = format_table(data, fmt=FORMAT_GENERATOR_COLS)
            for i, c in enumerate(header):
                if i > 0:
                    print(spacer, end='')
                print(colored(c, color='blue'), end='')
            print()
            print(
                colored('-' * (sum(len(s) for s in header) +
                               (len(header) - 1) * 2),
                        color='grey'))
            for d, r in zip(data, rows):
                for i, c in enumerate(r):
                    if i > 0:
                        print(spacer, end='')
                    if i == 2:
                        print(colored(c,
                                      color=ACTION_STATUS_COLOR.get(
                                          d['status'])),
                              end='')
                    else:
                        print(c, end='')
                print()

    def item_summary(self):
        data = call_rpc('item.summary')
        if current_command.json:
            print_result(data)
        else:
            print_result(data['sources'], name_value=['node', 'items'])
            print('total:', colored(data['items'], color='yellow'))
            print()

    def node_list(self, repl_svc, default_rpl):
        if default_rpl and repl_svc is None:
            repl_svc = DEFAULT_REPL_SERVICE
        if repl_svc is None:
            data = call_rpc('node.list')
            if not current_command.json:
                for d in data:
                    if d['remote']:
                        d['type'] = 'remote'
                    else:
                        d['type'] = 'local'
                    try:
                        info = d.pop('info')
                        d['version'] = info.get('version')
                        d['build'] = info.get('build')
                    except KeyError:
                        pass
            print_result(
                data,
                cols=['name', 'svc', 'type', 'online', 'version', 'build'])
        else:
            data = call_rpc('node.list', target=repl_svc)
            print_result(data,
                         cols=[
                             'name', 'timeout', 'compress|n=compr',
                             'ping_interval|n=ping int',
                             'reload_interval|n=rel.int', 'static', 'enabled',
                             'managed', 'online',
                             'link_uptime|n=link upt|f=round:0', 'version',
                             'build'
                         ])

    def node_append(self, i, repl_svc):
        call_rpc('node.append', dict(i=i), target=repl_svc)
        ok()

    def node_edit(self, i, repl_svc):

        def deploy_edited_node(cfg, i):
            call_rpc('node.deploy', dict(nodes=[cfg]), target=repl_svc)
            print(f'node re-deployed: {i}')
            print()

        config = call_rpc('node.get_config', dict(i=i), target=repl_svc)
        edit_config(config,
                    f'node|{i}|config',
                    deploy_fn=partial(deploy_edited_node, i=i))

    def node_export(self, i, repl_svc, output=None):
        configs = call_rpc('node.export', dict(i=i), target=repl_svc)
        c = len(configs['nodes'])
        if current_command.json:
            print_result(configs)
        else:
            import yaml
            dump = yaml.dump(configs, default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} node(s) exported')
                print()

    def node_test(self, i, repl_svc):
        call_rpc('node.test', dict(i=i), target=repl_svc)
        ok()

    def node_mtest(self, i, repl_svc):
        call_rpc('node.mtest', dict(i=i), target=repl_svc)
        ok()

    def node_remove(self, i, repl_svc):
        call_rpc('node.remove', dict(i=i), target=repl_svc)
        ok()

    def node_reload(self, i, repl_svc):
        call_rpc('node.reload', dict(i=i), target=repl_svc)
        print('reload request accepted')
        print()

    def node_deploy(self, repl_svc, file=None):
        import yaml
        nodes = yaml.safe_load(read_file(file).decode()).pop('nodes')
        call_rpc('node.deploy', dict(nodes=nodes), target=repl_svc)
        print(f'{len(nodes)} node(s) deployed')
        print()

    def node_undeploy(self, repl_svc, file=None):
        import yaml
        nodes = yaml.safe_load(read_file(file).decode()).pop('nodes')
        call_rpc('node.undeploy', dict(nodes=nodes), target=repl_svc)
        print(f'{len(nodes)} node(s) undeployed')
        print()

    def item_list(self, i, n=None):
        data = call_rpc(
            'item.list',
            dict(i=i, node=n),
        )
        print_result(data,
                     cols=[
                         'oid', 'status', 'value', 't|n=set time|f=time',
                         'node', 'connected', 'enabled'
                     ])

    def item_announce(self, i, n=None):
        data = call_rpc(
            'item.announce',
            dict(i=i, node=n),
        )
        ok()

    def item_export(self, i, output=None):
        c = 0
        configs = []
        for item in call_rpc('item.list', dict(i=i, node='.local')):
            c += 1
            configs.append(call_rpc('item.get_config', dict(i=item['oid'])))
        result = dict(items=configs)
        if current_command.json:
            print_result(result)
        else:
            import yaml
            dump = yaml.dump(result, default_flow_style=False)
            if output is None:
                print(dump)
            else:
                write_file(output, dump)
                print(f'{c} item(s) exported')
                print()

    def item_set(self, i, status, value):
        payload = dict(status=status, force=True)
        if value is not None:
            val = format_value(value)
            payload['value'] = val
        common.bus.send(
            'RAW/' + i.replace(':', '/', 1),
            elbus.client.Frame(msgpack.dumps(payload),
                               tp=elbus.client.OP_PUBLISH,
                               qos=1)).wait_completed()
        time.sleep(0.1)
        state = call_rpc('item.state', dict(i=i))[0]
        try:
            if state['status'] != status or (value is not None and
                                             state['value'] != val):
                err('FAILED')
            else:
                ok()
        except KeyError:
            err(f'{i} has no status')

    def item_deploy(self, file=None):
        import yaml
        items = yaml.safe_load(read_file(file).decode()).pop('items')
        call_rpc('item.deploy', dict(items=items))
        print(f'{len(items)} item(s) deployed')
        print()

    def item_undeploy(self, file=None):
        import yaml
        items = yaml.safe_load(read_file(file).decode()).pop('items')
        call_rpc('item.undeploy', dict(items=items))
        print(f'{len(items)} item(s) undeployed')
        print()

    def item_create(self, i):
        call_rpc('item.create', dict(i=i))
        ok()

    def item_destroy(self, i):
        call_rpc('item.destroy', dict(i=i))
        ok()

    def item_enable(self, i):
        call_rpc('item.enable', dict(i=i))
        ok()

    def item_edit(self, i):

        def deploy_edited_item(cfg, i):
            call_rpc('item.deploy', dict(items=[cfg]))
            print(f'item re-deployed: {i}')
            print()

        config = call_rpc('item.get_config', dict(i=i))
        edit_config(config,
                    f'item|{i}|config',
                    deploy_fn=partial(deploy_edited_item, i=i))

    def item_disable(self, i):
        call_rpc('item.disable', dict(i=i))
        ok()

    def item_state(self, i, full):
        data = call_rpc(
            'item.state',
            dict(i=i, full=full),
        )
        print_result(data,
                     cols=[
                         'oid', 'status', 'value', 'ieid',
                         't|n=set time|f=time', 'node', 'connected', 'act'
                     ])

    def item_slog(self, i, db_svc, time_start, time_end, time_zone, limit):
        time_start = prepare_time(time_start)
        time_end = prepare_time(time_end)
        data = call_rpc('state_log',
                        dict(i=i,
                             t_start=time_start,
                             t_end=time_end,
                             limit=limit),
                        target=db_svc)
        print_result(
            data,
            cols=[
                'oid', 'status', 'value',
                't|n=time|f=time{}'.format(f':{time_zone}' if time_zone else '')
            ])

    def item_history(self, i, db_svc, time_start, time_end, time_zone, limit,
                     prop, fill):
        time_start = prepare_time(time_start)
        time_end = prepare_time(time_end)
        precs = None
        if fill is not None:
            if ':' in fill:
                fill, precs = fill.split(':', 1)
                precs = int(precs)
        data = call_rpc('state_history',
                        dict(i=i,
                             t_start=time_start,
                             t_end=time_end,
                             fill=fill,
                             precision=precs,
                             prop=prop,
                             limit=limit),
                        target=db_svc)
        cols = []
        if prop == 'status' or prop is None:
            cols += ['status']
        if prop == 'value' or prop is None:
            cols += ['value']
        cols += [
            't|n=time|f=time{}'.format(f':{time_zone}' if time_zone else '')
        ]
        print_result(data, cols=cols)

    def item_watch(self, i, interval, rows, prop, chart_type):
        from . import charts
        import datetime
        vals = []
        to_clear = 0
        old_width = 0
        old_height = 0
        limit = rows
        limit_auto_set = rows is None

        def _append(label, value, limit):
            if chart_type == 'bar':
                vals.append((label, value))
            else:
                if isinstance(value, float):
                    vals.append(value)
                else:
                    try:
                        v = float(value)
                    except:
                        v = 0
                    vals.append(v)
            while len(vals) > limit:
                vals.pop(0)

        next_step = time.perf_counter() + interval
        try:
            while True:
                if chart_type == 'bar':
                    label = datetime.datetime.now().strftime('%T.%f')[:-5]
                else:
                    label = datetime.datetime.now().isoformat()
                width, height = get_term_size()
                if limit is None:
                    if chart_type == 'bar':
                        limit = height - 2
                    else:
                        limit = width - 12
                if width != old_width or height != old_height:
                    os.system('clear')
                    if chart_type == 'bar':
                        print(colored(i, color='yellow'))
                    old_width, old_height = width, height
                    if limit_auto_set:
                        if chart_type == 'bar':
                            limit = height - 2
                        else:
                            rows = height - 6
                elif chart_type == 'bar':
                    for t in range(to_clear):
                        sys.stdout.write('\033[F\033[K')
                elif chart_type == 'line':
                    os.system('clear')
                res = call_rpc('item.state', dict(i=i))
                for data in res:
                    if data['oid'] == i:
                        break
                else:
                    raise ValueError('no state returned')
                v = data.get(prop)
                if isinstance(v, str):
                    try:
                        v = int(v)
                    except:
                        try:
                            v = float(v)
                        except:
                            v = None
                if chart_type == 'bar':
                    _append(label, v, limit)
                    charts.plot_bar_chart(vals)
                    to_clear = len(vals)
                else:
                    _append(label, v, width)
                    max_value_width = 0
                    for z in vals:
                        x = len(str(z))
                        if x > max_value_width:
                            max_value_width = x
                    while len(vals) > width - (max_value_width if
                                               max_value_width > 7 else 7) - 5:
                        vals.pop(0)
                    max_value_width += 2
                    print(
                        f'{colored(i, color="yellow")} '
                        f'{colored(label, color="cyan")}: ',
                        end='')
                    if v is None:
                        print(colored('NaN', color='magenta'))
                    else:
                        print(colored(v, color='green', attrs='bold'))
                    charts.plot_line_chart(vals, rows, max_value_width)
                t = time.perf_counter()
                sleep_to = next_step - t
                if sleep_to > 0:
                    time.sleep(sleep_to)
                    next_step += interval
                else:
                    next_step = t
        except KeyboardInterrupt:
            return
        pass

    def lvar_set(self, i, status, value):
        params = dict(i=i)
        if status is not None:
            params['status'] = status
        if value is not None:
            params['value'] = format_value(value)
        call_rpc('lvar.set', params)
        ok()

    def lvar_reset(self, i):
        call_rpc('lvar.reset', dict(i=i))
        ok()

    def lvar_clear(self, i):
        call_rpc('lvar.clear', dict(i=i))
        ok()

    def lvar_toggle(self, i):
        call_rpc('lvar.toggle', dict(i=i))
        ok()

    def lvar_incr(self, i):
        val = call_rpc('lvar.incr', dict(i=i))
        print(val)

    def lvar_decr(self, i):
        val = call_rpc('lvar.decr', dict(i=i))
        print(val)

    def system_update(self):
        import shutil
        apt = shutil.which('apt')
        if apt:
            update = 'apt update'
            upgrade = 'apt -y upgrade'
            clean = 'apt -y clean'
        else:
            apk = shutil.which('apk')
            if apk:
                update = 'apk update'
                upgrade = 'apk upgrade'
                clean = None
            else:
                raise RuntimeError(
                    'Unsupported OS, please update the system manually')
        if update:
            xc(update,
               ps='Updating package indexes',
               verbose=current_command.debug)
        if upgrade:
            xc(upgrade, ps='Upgrading packages', verbose=current_command.debug)
        if clean:
            xc(clean, ps='Cleaning up', verbose=current_command.debug)
        xc('sync', ps='Flushing data to disk', verbose=current_command.debug)
        print('System updated')
        print()

    def system_reboot(self):
        import getch
        print('Reboot the system? (y/n) ', end='', flush=True)
        c = getch.getch()
        print()
        if c in ['y', 'Y']:
            warn('REBOOTING THE SYSTEM...')
            if os.system('reboot'):
                raise RuntimeError
            while True:
                time.sleep(1)

    def system_poweroff(self):
        import getch
        print('Power off the system? (y/n) ', end='', flush=True)
        c = getch.getch()
        print()
        if c in ['y', 'Y']:
            warn('POWERING OFF...')
            if os.system('poweroff'):
                raise RuntimeError
            while True:
                time.sleep(1)

    def version(self):
        from . import __version__
        info = get_node_svc_info()
        if current_command.json:
            d = {
                'cli_version': __version__,
                'build': info['build'],
                'version': info['version'],
            }
        else:
            d = [{
                'name': 'version',
                'value': info['version']
            }, {
                'name': 'build',
                'value': info['build']
            }, {
                'name': 'CLI version',
                'value': __version__
            }]
        ip = get_my_ip()
        if ip:
            if current_command.json:
                d['ip'] = ip
            else:
                d.append({'name': 'IP address', 'value': ip})
        print_result(d, need_header=False, name_value=False)

    def mirror_set(self, mirror_url):
        args = f'node mirror-set "{mirror_url}"'
        exec_cmd('eva-cloud-manager', args)
        self.venv_mirror_set(mirror_url=mirror_url)

    def mirror_update(self, dest, current_arch_only, force, repository_url):
        args = f'node mirror-update --timeout {current_command.timeout}'
        if dest:
            args += f' --dest "{dest}"'
        if repository_url:
            args = + f' --repository-url "{repository_url}"'
        if force:
            args += ' --force'
        if current_arch_only:
            args += ' --current-arch-only'
        exec_cmd('eva-cloud-manager', args)
        self.venv_mirror_update(dest=dest)
        print()
        self.mirror_info()

    def mirror_info(self):
        ip = get_my_ip()
        port = 7780
        if ip is None:
            ip = '<host/ip>'
        mirror_url = f'http://{ip}:{port}'
        print(f'Mirror URL: ' +
              colored(mirror_url, color='green', attrs=['bold']))
        print()
        print(f'EVA ICS update: -u http://{ip}:{port}/eva')
        print(f'pip-extra-options: "-i {mirror_url}/pypi/local'
              f' --trusted-host {ip}"')
        print()
        print('To automatically set the mirror config on'
              ' other nodes, execute there a command:')
        print()
        print(
            colored(f'    eva mirror set {mirror_url}',
                    color='green',
                    attrs=['bold']))
        print()
        print('Note: deploy the mirror service if not deployed.')
        print('If the mirror is not created yet, '
              'create it with "eva mirror update" command.')
        print('If the mirror service is set to different port, '
              'replace it in the URLs above.')



import sys
import os
from neotermcolor import colored

from .sharedobj import common, current_command
from .cli import CLI
from .ap import init_ap

banner = """     _______    _____       _______________
    / ____/ |  / /   |     /  _/ ____/ ___/
   / __/  | | / / /| |     / // /    \__ \\
  / /___  | |/ / ___ |   _/ // /___ ___/ /
 /_____/  |___/_/  |_|  /___/\____//____/"""

home_url = "www.eva-ics.com"
copyright = "(c) 2022 Bohemia Automation"

dir_eva_default = '/opt/eva4'


def launch():
    common.dir_eva = os.environ.get('EVA_DIR', dir_eva_default)
    if not os.path.exists(common.dir_eva):
        raise RuntimeError(
            f'EVA ICS directory not found. Consider installing EVA ICS either '
            f'in {dir_eva_default} or specifying EVA_DIR env variable')
    common.elbus_path = f'{common.dir_eva}/var/elbus.ipc'
    common.elbus_name = f'eva-shell-{os.getpid()}'

    common.cli = CLI()

    ap = init_ap()

    if len(sys.argv) > 1:
        ap.launch()
        exit(current_command.exit_code)
    else:
        lines = banner.split('\n')
        size = len(max(lines, key=len))
        pal = 3
        n = size // (pal - 1)
        for line in banner.split('\n'):
            chunks = [line[i:i + n] for i in range(0, len(line), n)]
            for i, ch in enumerate(chunks):
                if i > pal:
                    i = pal
                print(colored(ch, color=31 + i), end='')
            print()
        print()
        print(' ' + colored(home_url, color=24, attrs=['underline']), end=' ')
        print(colored(copyright, color=25))
        print()
        common.cli.version()
        common.interactive = True
        ap.interactive()
        print('Bye')



import os
from .tools import can_colorize, safe_print, get_term_size
from neotermcolor import colored


def plot_bar_chart(data):
    BAR = '❚'
    TICK = '▏'

    width, height = get_term_size()
    spaces = 8
    extra = 20 if can_colorize() else 0
    max_label_length = 0
    max_value_label_length = 0
    min_value = None
    max_value = None
    for label, value in data:
        if value is not None:
            if len(label) > max_label_length:
                max_label_length = len(label)
            if len(str(value)) > max_value_label_length:
                max_value_label_length = len(str(value))
            if min_value is None or value < min_value:
                min_value = value
            if max_value is None or value > max_value:
                max_value = value
    if max_value is None:
        max_value = 0
    if min_value is None:
        min_value = 0
    plot_width = width - spaces - max_label_length - max_value_label_length
    if max_value > 0:
        max_bar_len = max_value - min_value
    else:
        max_bar_len = abs(min_value - max_value)
    zoom = plot_width / max_bar_len if max_bar_len else 1
    for label, value in data:
        if value is None:
            line = colored('NaN', color='magenta')
        else:
            if min_value >= 0:
                line = BAR * int((value - min_value) * zoom)
                if not line:
                    line = f'{TICK if value else " "}{value}'
                else:
                    line += f' {value}'
                line = colored(line,
                               color='green' if value is not None else 'grey')
            elif max_value < 0:
                line = BAR * int((abs(max_value - value)) * zoom)
                if not line:
                    line = f'{value} {(TICK) if value else ""}'
                    line = line.rjust(plot_width + spaces + 1)
                else:
                    line = f'{value} ' + line
                    line = line.rjust(plot_width + spaces)
                line = colored(line,
                               color='yellow' if value is not None else 'grey')
            else:
                if value >= 0:
                    line = BAR * int(value * zoom) + f' {value}'
                    line = colored(
                        line, color='green' if value is not None else 'grey')
                    line = ' ' * (1 + int(
                        abs(min_value) * zoom + len(str(min_value)))) + line
                else:
                    bar = BAR * int(abs(value) * zoom)
                    if not bar:
                        bar = TICK
                    line = f'{value} ' + bar
                    line = line.rjust(
                        1 + int(abs(min_value) * zoom + len(str(min_value))))
                    line = colored(line, color='yellow')
        safe_print(
            f'{colored(label.rjust(max_label_length), color="cyan")}'
            f' {line}', extra)


def plot_line_chart(data, rows, max_label_width):
    # https://github.com/kroitor/asciichart/tree/master/asciichartpy
    # Copyright © 2016 Igor Kroitor
    from math import ceil, floor, isnan

    def _isnum(n):
        return not isnan(n)

    def colored(char, color):
        if not color:
            return char
        else:
            return color + char + "\033[0m"

    def _plot(series, cfg=None):
        if len(series) == 0:
            return ''

        if not isinstance(series[0], list):
            if all(isnan(n) for n in series):
                return ''
            else:
                series = [series]

        cfg = cfg or {}
        colors = cfg.get('colors', [None])
        minimum = cfg.get('min',
                          min(filter(_isnum, [j for i in series for j in i])))
        maximum = cfg.get('max',
                          max(filter(_isnum, [j for i in series for j in i])))
        default_symbols = ['┼', '┤', '╶', '╴', '─', '╰', '╭', '╮', '╯', '│']
        symbols = cfg.get('symbols', default_symbols)
        if minimum > maximum:
            raise ValueError('The min value cannot exceed the max value.')
        interval = maximum - minimum
        offset = cfg.get('offset', 3)
        height = cfg.get('height', interval)
        ratio = height / interval if interval > 0 else 1
        min2 = int(floor(minimum * ratio))
        max2 = int(ceil(maximum * ratio))

        def clamp(n):
            return min(max(n, minimum), maximum)

        def scaled(y):
            return int(round(clamp(y) * ratio) - min2)

        rows = max2 - min2

        width = 0
        for i in range(0, len(series)):
            width = max(width, len(series[i]))
        width += offset
        placeholder = cfg.get(
            'format',
            '{:' + str(8 if max_label_width < 8 else max_label_width) + '.2f} ')
        result = [[' '] * width for i in range(rows + 1)]
        # axis and labels
        for y in range(min2, max2 + 1):
            label = placeholder.format(maximum - ((y - min2) * interval /
                                                  (rows if rows else 1)))
            result[y - min2][max(offset - len(label), 0)] = label
            result[y - min2][offset - 1] = symbols[0] if y == 0 else symbols[
                1]  # zero tick mark
        # first value is a tick mark across the y-axis
        d0 = series[0][0]
        if _isnum(d0):
            result[rows - scaled(d0)][offset - 1] = symbols[0]
        for i in range(0, len(series)):
            color = colors[i % len(colors)]
            # plot the line
            for x in range(0, len(series[i]) - 1):
                d0 = series[i][x + 0]
                d1 = series[i][x + 1]
                if isnan(d0) and isnan(d1):
                    continue
                if isnan(d0) and _isnum(d1):
                    result[rows - scaled(d1)][x + offset] = colored(
                        symbols[2], color)
                    continue
                if _isnum(d0) and isnan(d1):
                    result[rows - scaled(d0)][x + offset] = colored(
                        symbols[3], color)
                    continue
                y0 = scaled(d0)
                y1 = scaled(d1)
                if y0 == y1:
                    result[rows - y0][x + offset] = colored(symbols[4], color)
                    continue
                result[rows - y1][x + offset] = colored(
                    symbols[5], color) if y0 > y1 else colored(
                        symbols[6], color)
                result[rows - y0][x + offset] = colored(
                    symbols[7], color) if y0 > y1 else colored(
                        symbols[8], color)
                start = min(y0, y1) + 1
                end = max(y0, y1)
                for y in range(start, end):
                    result[rows - y][x + offset] = colored(symbols[9], color)
        return '\n'.join([''.join(row).rstrip() for row in result])

    config = {'height': rows}
    if can_colorize():
        config['colors'] = ['\033[32m']
    print(_plot(data, cfg=config))



import time
from types import SimpleNamespace
import pytz

common = SimpleNamespace(dir_eva=None,
                         elbus_path=None,
                         elbus_name=None,
                         bus=None,
                         rpc=None,
                         cli=None,
                         interactive=False,
                         public_key=None,
                         TZ=pytz.timezone(time.tzname[0]))
current_command = SimpleNamespace(json=False,
                                  debug=False,
                                  timeout=5,
                                  exit_code=0)
